@@ -2,9 +2,12 @@
 # See LICENSE (https://gnu.org/licenses/lgpl-3.0.txt) for more information.
 #  AdvancedTag and TagCollection which represent tags and their data.
 
+from collections import OrderedDict
+
 import uuid
 
 from .constants import PREFORMATTED_TAGS, IMPLICIT_SELF_CLOSING_TAGS
+from .SpecialAttributes import SpecialAttributesDict, StyleAttribute
 
 def uniqueTags(tagList):
     '''
@@ -152,7 +155,7 @@ class TagCollection(list):
             @return - a single tag matching the id, or None if none found
         '''
         for tag in self:
-            if tag.getId() == _id:
+            if tag.id == _id:
                 return tag
             for subtag in tag.children:
                 tmp = subtag.getElementById(_id)
@@ -201,7 +204,6 @@ class TagCollection(list):
         return ret
         
 
-
 class AdvancedTag(object):
     '''
         AdvancedTag - Represents a Tag. Used with AdvancedHTMLParser to create a DOM-model
@@ -224,22 +226,19 @@ class AdvancedTag(object):
         if isSelfClosing is False and tagName in IMPLICIT_SELF_CLOSING_TAGS:
             isSelfClosing = True
 
-        self.attributes = {}
-        if attrList is not None:
-            for key, value in attrList:
-                self.attributes[key.lower()] = value
-
+        self.attributes = SpecialAttributesDict(self)
         self.text = ''
         self.blocks = ['']
+        self.classNames = []
+        self.className = ''
+        self.style = StyleAttribute('')
 
         self.isSelfClosing = isSelfClosing
 
-        if 'class' in self.attributes:
-            self.className = self.attributes['class']
-            self.classNames = [x for x in self.attributes['class'].split(' ') if x]
-        else:
-            self.className = ''
-            self.classNames = []
+        if attrList is not None:
+            for key, value in attrList:
+                key = key.lower()
+                self.attributes[key] = value
 
         self.children = []
 
@@ -248,6 +247,14 @@ class AdvancedTag(object):
 
         self.indent = ''
 
+    def __setattr__(self, name, value):
+        if name == 'style' and not isinstance(value, StyleAttribute):
+            value = StyleAttribute(value)
+        try:
+            return object.__setattr__(self, name,  value)
+        except AttributeError:
+            raise AttributeError('Cannot set property %s. Use setAttribute?' %(name,))
+
     def appendText(self, text):
         '''
             appendText - append some inner text
@@ -278,6 +285,8 @@ class AdvancedTag(object):
     def appendChild(self, child):
         '''
             appendChild - Append a child to this element.
+
+            @param child <AdvancedTag> - Append a child element to this element
         '''
     
         child.parentNode = self
@@ -304,10 +313,107 @@ class AdvancedTag(object):
             return None
 
     removeNode = removeChild
-            
+
+    def insertBefore(self, child, beforeChild):
+        '''
+            insertBefore - Inserts a child before @beforeChild
+
+            child - Child to insert
+            beforeChild - Child  to insert before. if None, will  be appended
+
+        '''
+        if beforeChild is None:
+            return self.appendChild(child)
+
+        try:
+            childrenIdx = self.children.index(beforeChild)
+            blocksIdx =  self.blocks.index(beforeChild)
+            self.children = self.children[:childrenIdx] + [child] + self.children[childrenIdx:]
+            self.blocks = self.blocks[:blocksIdx] + [child] + self.blocks[blocksIdx:]
+        except ValueError:
+            raise ValueError('Provided "beforeChild" is not a child of element, cannot insert.')
+
+    def insertAfter(self, child, afterChild):
+        '''
+            insertAfter - Inserts a child after @afterChild
+
+            child - Child to insert
+            afterChild - Child  to insert after. if None, will  be appended
+
+        '''
+        if afterChild is None:
+            return self.appendChild(child)
+
+        try:
+            childrenIdx = self.children.index(afterChild)
+            blocksIdx =  self.blocks.index(afterChild)
+            self.children = self.children[:childrenIdx+1] + [child] + self.children[childrenIdx+1:]
+            self.blocks = self.blocks[:blocksIdx+1] + [child] + self.blocks[blocksIdx+1:]
+        except ValueError:
+            raise ValueError('Provided "afterChild" is not a child of element, cannot insert.')
+
+    @property
+    def nextSibling(self):
+        '''
+            nextSibling - Returns the next sibling.  This could be text or an element. use nextSiblingElement to ensure element
+        '''
+        if not self.parentNode:
+            return None
+        myBlockIdx = self.parentNode.blocks.index(self)
+        if myBlockIdx == len(self.parentNode.blocks):
+            return None
+        return self.parentNode.blocks[myBlockIdx+1]
+
+    @property
+    def nextSiblingElement(self):
+        '''
+            nextSiblingElement - Returns the next sibling  that is an element. 
+        '''
+        if not self.parentNode:
+            return None
+        myElementIdx = self.parentNode.children.index(self)
+        if myElementIdx == len(self.parentNode.children):
+            return None
+        return self.parentNode.children[myElementIdx+1]
+        
+    @property
+    def previousSibling(self):
+        '''
+            previousSibling - Returns the previous sibling.  This could be text or an element. use previousSiblingElement to ensure element
+        '''
+        if not self.parentNode:
+            return None
+        myBlockIdx = self.parentNode.blocks.index(self)
+        if myBlockIdx == 0:
+            return None
+        return self.parentNode.blocks[myBlockIdx-1]
+
+    @property
+    def previousSiblingElement(self):
+        '''
+            previousSiblingElement - Returns the previous  sibling  that is an element. 
+        '''
+        if not self.parentNode:
+            return None
+        myElementIdx = self.parentNode.children.index(self)
+        if myElementIdx == 0:
+            return None
+        return self.parentNode.children[myElementIdx-1]
+        
+
     def getChildren(self):
         '''
-            getChildren - returns child nodes
+            getChildren - returns child nodes as a searchable TagCollection.
+
+                @return - TagCollection of the immediate children to this tag.
+        '''
+        return TagCollection(self.children)
+
+    def getChildren(self):
+        '''
+            getChildren - returns child nodes as a searchable TagCollection.
+
+                @return - TagCollection of the immediate children to this tag.
         '''
         return TagCollection(self.children)
 
@@ -333,7 +439,7 @@ class AdvancedTag(object):
     @property
     def childNodes(self):
         '''
-            childNodes - returns child nodes
+            childNodes - returns immediate child nodes as a TagCollection
 
             @return - TagCollection of child nodes
         '''
@@ -452,13 +558,21 @@ class AdvancedTag(object):
 
     def setAttribute(self, attrName, attrValue):
         '''
-            setAttribute - Sets an attribute. Do not use this for classname, use addClass/removeClass. Attribute names are all lowercase.
+            setAttribute - Sets an attribute. Be wary using this for classname, maybe use addClass/removeClass. Attribute names are all lowercase.
         
             @param attrName <str> - The name of the attribute
             @param attrValue <str> - The value of the attribute
         '''
         self.attributes[attrName] = attrValue
 
+    def setAttributes(self, attributesDict):
+        '''
+            setAttributes - Sets  several attributes at once, using a dictionary of attrName : attrValue
+
+            @param  attributesDict - <str:str> - New attribute names -> values
+        '''
+        self.attributes.update(attributesDict)
+
     def hasAttribute(self, attrName):
         '''
             hasAttribute - Checks for the existance of an attribute. Attribute names are all lowercase.
@@ -467,8 +581,22 @@ class AdvancedTag(object):
                 
                 @return <bool> - True or False if attribute exists by that name
         '''
+        attrName = attrName.lower()
         return bool(attrName in self.attributes)
 
+    def removeAttribute(self, attrName):
+        '''
+            removeAttribute - Removes an attribute, by name.
+            
+            @param attrName <str> - The attribute name
+
+        '''
+        attrName = attrName.lower()
+        try:
+            del self.attributes[attrName]
+        except KeyError:
+            pass
+
     def hasClass(self, className):
         '''
             hasClass - Test if this tag has a paticular class name
@@ -498,7 +626,73 @@ class AdvancedTag(object):
             return className
 
         return None
-   
+
+    def getStyleDict(self):
+        '''
+            getStyleDict - Gets a dictionary of style attribute/value pairs.
+
+            @return - OrderedDict of "style" attribute.
+        '''
+        styleStr = (self.getAttribute('style') or '').strip()
+        styles = styleStr.split(';') # Won't work for strings containing semicolon..
+        styleDict = OrderedDict()
+        for item in styles:
+            try:
+                splitIdx = item.index(':')
+                name = item[:splitIdx].strip().lower()
+                value = item[splitIdx+1:].strip()
+                styleDict[name] = value
+            except:
+                continue
+
+        return styleDict
+
+
+    def getStyle(self, styleName):
+        '''
+            getStyle - Gets the value of a style paramater, part of the "style" attribute
+
+            @param styleName - The name of the style
+
+            @return - String of the value of the style. '' is no value.
+        '''
+        return getattr(self.style, styleName.lower())
+        
+
+    def setStyle(self, styleName, styleValue):
+        '''
+            setStyle - Sets a style param. Example: "display", "block"
+
+                If you need to set many styles on an element, use setStyles instead. 
+                It takes a dictionary of attribute, value pairs and applies it all in one go (faster)
+
+                To remove a style, set its value to empty string.
+                When all styles are removed, the "style" attribute will be nullified.
+
+            @param styleName - The name of the style element
+            @param styleValue - The value of which to assign the style element
+
+            @return - String of current value of "style" after change is made.
+        '''
+        setattr(self.style, styleName, styleValue)
+
+    def setStyles(self, styleUpdatesDict):
+        '''
+            setStyles - Sets one or more style params. 
+                This all happens in one shot, so it is much much faster than calling setStyle for every value.
+
+                To remove a style, set its value to empty string.
+                When all styles are removed, the "style" attribute will be nullified.
+
+            @param styleUpdatesDict - Dictionary of attribute : value styles.
+
+            @return - String of current value of "style" after change is made.
+        '''
+        for newName, newValue in styleUpdatesDict.items():
+            setattr(self.style, newName, newValue)
+
+        return self.style
+
 
     def __str__(self):
         '''
